/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").  
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at sunpinyin/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at sunpinyin/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */
/*
 * Copyright 2007 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <deque>

#include "pytrie.h"

const unsigned char CPinyinTrie::SYLLABLE_BREAKER;

const CPinyinTrie::TNode*
CPinyinTrie::transfer(const TNode* pnode, const unsigned char* str) const
{
    if (str) {
        for (; *str != 0 && pnode != NULL; ++str)
            pnode = transfer(pnode, *str);
    }
    return pnode;
}

const CPinyinTrie::TNode*
CPinyinTrie::transfer(const TNode* pnode, const TWCHAR* wstr) const
{
    if (wstr) {
        for (; *wstr != WCH_NULL && pnode != NULL; ++wstr)
            pnode = transfer(pnode, unsigned(*wstr));
    }
    return pnode;
}

const CPinyinTrie::TNode*
CPinyinTrie::transfer(const TNode* pnode, const unsigned char* str, int nlen) const
{
    if (str && nlen > 0) {
        for (; nlen>0 && *str != 0 && pnode != NULL; --nlen)
            pnode = transfer(pnode, *str++);
    }
    return pnode;
}

const CPinyinTrie::TNode*
CPinyinTrie::transfer(const TNode* pnode, const TWCHAR* wstr, int nlen) const
{
    if (wstr && nlen > 0) {
        for (; nlen>0 && *wstr != WCH_NULL && pnode != NULL; --nlen)
            pnode = transfer(pnode, unsigned(*wstr++));
    }
    return pnode;
}

bool
CPinyinTrie::isValid(const TNode* pnode, bool allowNonComplete, bool allowGBK)
{
    if ((pnode != NULL) && (allowGBK || pnode->m_bGBK == 0) && (pnode->transfer(SYLLABLE_BREAKER) != 0))
        return (allowNonComplete || (pnode->m_bFullSyllableTransfer == 1));
    return false;
}

unsigned int
CPinyinTrie::TNode::transfer(unsigned char c) const
{
    unsigned int s = 0, t = m_nTransfer;
    const TTransUnit* ptrans = getTrans();
    while (s < t) {
        int m = (s+t)/2;
        if (ptrans[m].m_Char == c)
            return ptrans[m].m_Offset;
        if (ptrans[m].m_Char < c)
            s = m + 1;
        else
            t = m;
    }
    return 0;
}

int
CPinyinTrie::lengthAt(unsigned int idx) const
{
    if (idx < getWordCount() -1 ) {
       return (m_words[idx+1] - m_words[idx]) - 1;
    } else if (idx == getWordCount() - 1) {
        return (((TWCHAR*)(m_mem+m_Size))-m_words[idx])-1;
    }
    return 0;
}

void
CPinyinTrie::print(FILE *fp) const
{
    std::string prefix;
    print(getRootNode(), prefix, fp);
}

void
CPinyinTrie::print(const TNode* pRoot, std::string& prefix, FILE *fp) const
{
    static char buf[1024];
    if (pRoot->m_nWordId > 0) {
        fprintf(fp, "%s", prefix.c_str());
        unsigned int sz = pRoot->m_nWordId;
        const TWordIdInfo *pwids = pRoot->getWordIdPtr();
        for (unsigned int i = 0; i < sz; ++i) {
            unsigned int id = pwids[i].m_id;
            const TWCHAR *pw = operator[](id);
            int len = WCSLEN(pw);
            if (len != lengthAt(id)) {
                printf(" (lengthAt %d error) ", id);
            }
            WCSTOMBS(buf, pw, 1024);
            fprintf(fp, " %s", buf);
            if (pwids[i].m_bSeen == 0)
                fprintf(fp, "[x]");
            else
                fprintf(fp, "[o]");
        }
        fprintf(fp, "\n");
    }
    unsigned int sz = pRoot->m_nTransfer;
    const TTransUnit* ptrans = pRoot->getTrans();
    for (unsigned int i = 0; i < sz; ++i) {
        unsigned char c = (unsigned char)(ptrans[i].m_Char);
        const TNode *pch = transfer(pRoot, c);
        prefix += c;
        print(pch, prefix, fp);
        prefix.resize(prefix.size()-1);
    }
}

bool
CPinyinTrie::load(const char* fileName)
{
    bool suc = false;
    FILE *fp = fopen(fileName, "rb");
    if (fp != NULL) {
        suc = load(fp);
        fclose(fp);
    }
    return suc;
}

unsigned int
CPinyinTrie::getSimbolId(const TWCHAR* wstr)
{
    std::map<wstring, unsigned>::const_iterator it;

    it = m_SimbolMap.find(wstring(wstr));
    if (it != m_SimbolMap.end())
        return it->second;
    return 0;
}

unsigned int
CPinyinTrie::getSimbolId(const wstring & wstr)
{
    std::map<wstring, unsigned>::const_iterator it;

    it = m_SimbolMap.find(wstr);
    if (it != m_SimbolMap.end())
        return it->second;
    return 0;
}

void
CPinyinTrie::free(void)
{
    if (m_mem) {
        delete []m_mem;
        m_mem = NULL;
    }
    if (m_words) {
        delete []m_words;
        m_words = NULL;
    }
    m_SimbolMap.clear();
}

bool
CPinyinTrie::load(FILE* fp)
{
    free();

    bool suc = (fread(&m_Size, sizeof(unsigned int), 1, fp) == 1);

    suc = suc && ((m_mem = new char [m_Size]) != NULL);

    suc = suc && (fread(m_mem, m_Size, 1, fp) == 1);

    suc = suc && ((m_words = new TWCHAR*[getWordCount()]) != NULL);

    if (suc) {
        TWCHAR *p = (TWCHAR *)(m_mem + getStringOffset());
        for (int i=0, sz=getWordCount(); i < sz; ++i) {
            m_words[i] = p;
            while (*p++)
                ;
        }
        for (unsigned i=1; i < 100; ++i) {
            if (*m_words[i] != WCH_NULL && *m_words[i] != WCH_LESSTHAN)
                m_SimbolMap[wstring(m_words[i])] = i;
        }
    }
    return suc;
}
